from typing import Dict, Optional, Union
import logging

from aiohttp import ClientSession, ClientTimeout, ClientError

from scrapio.structures.proxies import AbstractProxyManager
from scrapio.requests.client import AbstractClient
from scrapio.requests.response import from_aiohttp_response
from scrapio.utils.helpers import get_proxy_from_manager


_DEFAULT_HEADERS = {
        "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8",
        "Accept-Encoding": "gzip, deflate, br",
        "Accept-Language": "en-US,en;q=0.9",
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/68.0.3440.106 Safari/537.36",
    }


def _get_headers(additional_headers: Optional[Dict[str, str]]):
    if additional_headers and isinstance(additional_headers, dict):
        return {**_DEFAULT_HEADERS, **additional_headers}
    return _DEFAULT_HEADERS


def _build_client_session(timeout: Union[float, int], **kwargs) -> ClientSession:
    aiohttp_timeout_rules = kwargs.get("aiohttp_timeout_rules")
    if aiohttp_timeout_rules and isinstance(aiohttp_timeout_rules, dict):
        aiohttp_timeout = ClientTimeout(**aiohttp_timeout_rules)
        return ClientSession(headers=_get_headers(kwargs.get("headers")), timeout=aiohttp_timeout)
    else:
        return ClientSession(headers=_get_headers(kwargs.get("headers")), conn_timeout=float(timeout))


class DefaultClient(AbstractClient):

    def __init__(self, timeout, **kwargs):
        self.session = _build_client_session(timeout, **kwargs)

    async def get_request(self, url: str, proxy_manager: Optional[AbstractProxyManager]):
        proxy = await get_proxy_from_manager(proxy_manager)
        async with self.session.get(url, proxy=proxy) as resp:
            try:
                await resp.read()
            except ClientError:
                logger = logging.getLogger("ScrapIO")
                logger.warning("ClientError for URL: {}".format(url))
            except Exception as e:
                logger = logging.getLogger("ScrapIO")
                logger.warning("Unexpected error for URL: {}".format(e))
            else:
                return await from_aiohttp_response(resp)

    async def close(self):
        await self.session.close()





from typing import List
from urllib.parse import urlparse, urljoin, urldefrag

from scrapio.requests.response import Response
import lxml.html as lh

from scrapio.structures.filtering import URLFilter


def link_extractor(
    response: Response, url_filter: URLFilter, defrag: bool
) -> (str, List[str]):
    html = response.body
    req_url = response.url
    dom = lh.fromstring(html)
    found_urls = []
    for href in dom.xpath("//a/@href"):
        url = urljoin(str(req_url), href)
        if defrag:
            url = urldefrag(url)[0]
        netloc = urlparse(url).netloc
        can_crawl = url_filter.can_crawl(netloc, url)
        if can_crawl:
            found_urls.append(url)
    return response, found_urls



import asyncio
import logging
from typing import Union, List, Any


from aiohttp import ClientTimeout

from scrapio.parsing.links import link_extractor
from scrapio.requests.default_client import DefaultClient
from scrapio.requests.response import Response
from scrapio.structures.queues import WorkQueue
from scrapio.structures.proxies import AbstractProxyManager
from scrapio.structures.rate_limiter import RateLimiter
from scrapio.structures.filtering import URLFilter

__all__ = ["BaseCrawler"]


class BaseCrawler:
    def __init__(
        self,
        start_url: Union[List[str], str],
        max_crawl_size=None,
        timeout=30,
        verbose=True,
        logger_level=logging.WARN,
        **kwargs
    ):
        self._start_url = start_url
        self._client = DefaultClient(timeout, **kwargs)
        self._client_timeout: Union[None, ClientTimeout] = None
        self._timeout: Union[int, None] = None

        self._proxy_manager: Union[None, AbstractProxyManager] = (
            kwargs.get("proxy_manager")(**kwargs)
            if kwargs.get("proxy_manager")
            else None
        )
        self._url_filter = self._set_url_filter(start_url, **kwargs)
        self._queue = WorkQueue(
            max_crawl_size,
            start_url,
            seen_url_handler=kwargs.get("seen_url_handler", None),
        )
        logging.basicConfig(level=logger_level, format="%(message)s")
        self._logger = kwargs.get("logger", logging.getLogger("Scraper"))
        self._client_timeout = self._setup_timeout_rules(timeout)

        self.verbose = verbose
        self._rate_limiter = (
            RateLimiter(kwargs.get("rate_limit")) if kwargs.get("rate_limit") else None
        )

    @staticmethod
    def _set_url_filter(start_url, **kwargs) -> URLFilter:
        custom_filter = kwargs.get("custom_filter")
        if custom_filter and issubclass(custom_filter, URLFilter):
            return custom_filter(
                start_url,
                kwargs.get("additional_rules", []),
                kwargs.get("follow_robots", True),
                kwargs.get("defragment_urls", True),
            )
        return URLFilter(
            start_url,
            kwargs.get("additional_rules", []),
            kwargs.get("follow_robots", True),
            kwargs.get("defragment_urls", True),
        )

    @staticmethod
    def _setup_timeout_rules(timeout: Union[float, int], **kwargs) -> ClientTimeout:
        if kwargs.get("client_timeout_rules") and isinstance(
            "client_timeout_rules", dict
        ):
            rules = kwargs.get("client_timeout_rules")
            return ClientTimeout(**rules)
        return ClientTimeout(total=float(timeout))

    def _get_best_event_loop(self):
        try:
            import uvloop
        except ImportError:
            self._logger.info(
                "Uvloop not found reverting to base event loop implementation"
            )
            return asyncio.get_event_loop()
        else:
            asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
            return asyncio.get_event_loop()

    def parse_result(self, response: Response) -> Any:
        raise NotImplementedError

    async def save_results(self, result):
        raise NotImplementedError

    async def _make_requests(self, consumer: int, url: str):
        try:
            self._logger.info("Coroutine: {}, Requesting URL: {}".format(consumer, url))
            if self._rate_limiter:
                await self._rate_limiter.limited(url)
            resp = await self._client.get_request(url, self._proxy_manager)
            await asyncio.sleep(0.001)
            await self._parse_response(consumer, resp)
        except Exception as e:
            self._logger.warning(
                "Coroutine: {}, Encountered exception: {}".format(consumer, e)
            )
            raise e
        finally:
            self._queue.task_done()

    async def _parse_response(self, consumer: int, response: Response) -> None:
        defrag = self._url_filter.defragment
        try:
            response, links = link_extractor(response, self._url_filter, defrag)
            parsed_data = self.parse_result(response)
            await self.save_results(parsed_data)
            for link in links:
                await self._queue.put_url(link)
        except Exception as e:
            self._logger.warning(
                "Coroutine: {}, Encountered exception: {}".format(consumer, e)
            )

    async def _process(self, consumer: int):
        while True:
            try:
                job = await self._queue.get_job()
                await self._make_requests(consumer, job)
            except asyncio.CancelledError:
                return

    async def _crawl(self, workers):
        workers = [asyncio.Task(self._process(i)) for i in range(workers)]
        await self._queue.join()
        for worker in workers:
            worker.cancel()

    async def _close(self):
        await self._client.close()

    def run_crawler(self, workers: int) -> None:
        loop = asyncio.get_event_loop()
        try:
            loop.run_until_complete(self._crawl(workers))
        except KeyboardInterrupt:
            logging.info("Shutting down - received keyboard interrupt")
        finally:
            loop.run_until_complete(self._close())
            loop.close()



from typing import Dict, Optional, Any

from aiohttp import ClientResponse


class Response:

    __slots__ = [
        'url',
        'status',
        'headers',
        'body',
        'raw_response'
    ]

    def __init__(self):
        self.url: Optional[str] = None
        self.status: Optional[int] = None
        self.headers: Optional[Dict] = None
        self.body: Optional[str] = None
        self.raw_response: Optional[Any] = None

    def __repr__(self):
        return f"<Response: {self.url} {self.status}>"


async def from_aiohttp_response(client_response: ClientResponse) -> Response:
    resp = Response()
    resp.url = client_response.url
    resp.status = client_response.status
    resp.headers = client_response.headers
    bytes = await client_response.read()
    resp.body = bytes.decode("utf-8", errors='ignore')
    resp.raw_response = client_response
    return resp



from abc import abstractmethod, ABC


class AbstractUrlSet(ABC):
    @abstractmethod
    def put(self, url: str) -> None:
        ...

    @abstractmethod
    def __contains__(self, item):
        ...



from scrapio.url_set.abstract_set import AbstractUrlSet


class SetContainer(AbstractUrlSet):

    __slots__ = [
        '_seen'
    ]

    def __init__(self):
        self._seen = set()

    def put(self, url: str):
        self._seen.add(url)

    def __contains__(self, item):
        return item in self._seen



from abc import ABCMeta, abstractmethod
from typing import Optional

from scrapio.structures.proxies import AbstractProxyManager


class AbstractClient(metaclass=ABCMeta):

    @abstractmethod
    async def get_request(self, url: str, proxy_manager: Optional[AbstractProxyManager]):
        ...



from urllib.parse import urlparse
from typing import List

from scrapio.url_set.abstract_set import AbstractUrlSet


class TrieContainer(AbstractUrlSet):

    __slots__ = (
        'trie',
        '_end_symbol'
    )

    def __init__(self):
        self.trie = dict()
        self._end_symbol = "END"

    def make_parts(self, url: str) -> List[str]:
        parsed = urlparse(url)
        parts = [parsed.scheme, parsed.netloc]
        for item in parsed.path.split('/'):
            parts.append(item)
        return parts

    def __contains__(self, item):
        sub_trie = self.trie
        parts = self.make_parts(item)
        for char in parts:
            if char in sub_trie:
                sub_trie = sub_trie[char]
            else:
                return False
        else:
            if self._end_symbol in sub_trie:
                return True
            else:
                return False

    def put(self, url: str) -> None:
        if url in self:
            return
        temp_trie = self.trie
        parts = self.make_parts(url)
        for char in parts:
            if char in temp_trie:
                temp_trie = temp_trie[char]
            else:
                temp_trie = temp_trie.setdefault(char, {})
        temp_trie[self._end_symbol] = self._end_symbol



import asyncio
from typing import Union, List
import enum

from scrapio.url_set.set_container import SetContainer
from scrapio.url_set.abstract_set import AbstractUrlSet
from scrapio.url_set.trie_container import TrieContainer


class JobType(enum.Enum):
    Crawl = 1
    Parse = 2


class WorkQueue:

    __slots__ = [
        "_seen_urls",
        "_active_jobs",
        "_seen_semaphore",
        "_max_crawl_size",
        "_page_count",
        "_queue",
    ]

    def __init__(
        self,
        max_crawl_size: Union[int, None],
        seed_urls: Union[List[str], str],
        seen_url_handler: AbstractUrlSet = None,
    ):
        self._seen_urls: AbstractUrlSet = seen_url_handler or TrieContainer()
        self._active_jobs = 0
        self._max_crawl_size = max_crawl_size
        self._page_count = 0
        self._queue = asyncio.Queue()
        self.seed_queue(seed_urls)

    def seed_queue(self, seed_urls: Union[List[str], str]):
        if isinstance(seed_urls, str):
            self._queue.put_nowait(seed_urls)
        elif isinstance(seed_urls, (set, list)):
            for item in seed_urls:
                self._queue.put_nowait(item)

    async def get_job(self) -> str:
        return await self._queue.get()

    async def put_url(self, url):
        if url not in self._seen_urls:
            if self._max_crawl_size and self._page_count < self._max_crawl_size:
                await self._queue.put(url)
            elif self._max_crawl_size is None:
                await self._queue.put(url)
            self._page_count += 1
        self._seen_urls.put(url)

    def task_done(self):
        self._queue.task_done()

    async def join(self):
        return await self._queue.join()
